/*
 * wlanmodel/NodeManager.cc
 *
 * Copyright (C) 2008-2010 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include "NodeManager.h"

/** Register as simulation module */
Define_Module(NodeManager);

/**
 * Constructor
 */
NodeManager::NodeManager()
{
}

/**
 * Destructor
 */
NodeManager::~NodeManager()
{
}

/**
 * Check if node is registered
 */
cModule*
NodeManager::isRegistered(const host_id *hostId)
{
	cModule *host = NULL;
	cModuleType *moduleType = cModuleType::find("VirtualHost");
//	ev << "isRegistered(): id = " << std::string((char *) hostId) << "\n";				//TODO: only for debug!
	for (SubmoduleIterator submod(getParentModule()); !submod.end(); submod++)
	{
		host = submod();
		if (moduleType == host->getModuleType())		// only check VirtualHosts
		{
			std::string id = host->par("hostId");
			if ((std::string((char *) hostId).compare(id)) == 0)
			{
//				ev << "                host = " << host << "\n";						//TODO: only for debug!
				return host;
			}
		}
	}
//	ev << "host = NULL ==> unknown host!\n";											//TODO: only for debug!
    return NULL;
}

/**
 * Register node at NodeManager
 */
void
NodeManager::addNode(registrationMsg *regMsg)
{
	int port = 0;
	char ipAddr[40];
	int registeredHosts = par("numRegisteredHosts");

	// extract host information
    host_id *hostId = (host_id *) &((regMsg->header).id);
	registrationInfo *regInfo = &(regMsg->info);
	std::string hostName = (char *) &(regInfo->name);
	if (regInfo->addr.ss_family == AF_INET) 				// we have IPv4
	{
		inet_ntop(AF_INET,
				  &((struct sockaddr_in *) &regInfo->addr)->sin_addr,
				  ipAddr, 40);
		port = ntohs(((struct sockaddr_in *) &regInfo->addr)->sin_port);
	}
	if (regInfo->addr.ss_family == AF_INET6)				// we have IPv6
	{
		inet_ntop(AF_INET,
				  &((struct sockaddr_in6 *) &regInfo->addr)->sin6_addr,
				  ipAddr, 40);
		port = ntohs(((struct sockaddr_in6 *) &regInfo->addr)->sin6_port);
	}

	ev << getFullPath() << ": registering host " << hostName;
	ev << " (id: " << std::string((char *) hostId) << ")";
	ev << ", " << ipAddr << ":" << port;
	ev << ", Vifs: " << regInfo->numVif << "\n";

	// create VirtualHost module
	cModuleType *moduleType = cModuleType::find("VirtualHost");
	cModule *virtualHost = moduleType->create("host", getParentModule(), registeredHosts + 1, registeredHosts);

	// check number of virtual interfaces
	int numVif = regInfo->numVif;
	if (numVif > MAX_NUM_VIF)
	{
		numVif = MAX_NUM_VIF;
		ev << getFullPath() << ".addNode(): MAX_NUM_VIF = " << MAX_NUM_VIF << "; ";
		ev << "only registering " << MAX_NUM_VIF << " of " << numVif << "interfaces\n";
	}

	// set parameters
	virtualHost->par("hostId") = std::string((char *) hostId);
	virtualHost->par("hostName") = hostName;
	virtualHost->par("ipAddress") = ipAddr;
	virtualHost->par("port") = port;
	virtualHost->par("numVif") = numVif;
	virtualHost->finalizeParameters();
	virtualHost->buildInside();

	for (int i = 0; i < numVif; i++)
	{
		MACAddress *macAddr = new MACAddress();
		macAddr->setAddressBytes((unsigned char *) &(regInfo->vifData[i].vifMac));

		cModule *wlan = check_and_cast<cModule *>(virtualHost->getSubmodule("ath", i));
		Ieee80211Mac *mac = check_and_cast<Ieee80211Mac *>(wlan->getSubmodule("mac"));
		mac->par("address") = macAddr->str();
	}

	// initialize submodules
	virtualHost->callInitialize();

	// set virtual interface configuration
	VifBackend *handler = check_and_cast<VifBackend *>(virtualHost->getSubmodule("vifBackend"));
	handler->initializeVifParameters(regMsg->info.vifData);

	virtualHost->scheduleStart(simTime());

	// update NodeManager
	par("numRegisteredHosts") = registeredHosts + 1;
	handleParameterChange("numRegisteredHosts");

	// acknowledge registration
	unsigned int msgId = regInfo->msgId;
	handler->sendAck(msgId);
}

void
NodeManager::removeNode(cModule *host)
{
	std::string hostName = host->par("hostName");
	std::string hostId = host->par("hostId");
	ev << getFullPath() << ": removing host " << hostName;
	ev << " (id: " << hostId << ")\n";

	ChannelControl *cc;
	cc = ChannelControl::get();
	cc->removeHost(host);
//	host->callFinish();
	host->deleteModule();

	// update NodeManager
	int registeredHosts = par("numRegisteredHosts");
	par("numRegisteredHosts") = registeredHosts - 1;
	handleParameterChange("numRegisteredHosts");
}

void
NodeManager::handleParameterChange(const char *parname)
{
	// do nothing
}


